/*
 * Receive wall messages and pass them to a notification program via stdin.
 *
 * Copyright (C) 2014  Simon Ruderich
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <config.h>

#include <assert.h>
#include <fcntl.h>
#include <limits.h>
#include <pwd.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#ifdef USE_UTEMPTER
# include <utempter.h>
#endif
#ifdef USE_UTMPX
# include <utmpx.h>
#endif


static void sig_handler(int signal) {
    (void)signal;
}
static void setup_signals(void) {
    struct sigaction action;

    memset(&action, 0, sizeof(action));
    sigemptyset(&action.sa_mask);
    action.sa_handler = sig_handler;

    /* Handle all important signals which might be sent to us so we break out
     * of the read()-loop below and can perform our cleanup. */
    sigaction(SIGHUP,  &action, NULL);
    sigaction(SIGINT,  &action, NULL);
    sigaction(SIGQUIT, &action, NULL);
    sigaction(SIGUSR1, &action, NULL);
    sigaction(SIGUSR2, &action, NULL);

    /* Collect zombies automatically without having to call waitpid(2). */
    signal(SIGCHLD, SIG_IGN);
}

static int open_tty(void) {
    int ptm;
    const char *name;

    ptm = posix_openpt(O_RDWR);
    if (ptm < 0) {
        return -1;
    }
    if (grantpt(ptm) != 0) {
        return -1;
    }

    /* Prevent write access for other users so they can't use wall to send
     * messages to this program. */
    name = ptsname(ptm);
    if (!name) {
        return -1;
    }
    if (chmod(name, S_IRUSR | S_IWUSR) != 0) {
        return -1;
    }

    if (unlockpt(ptm) != 0) {
        return -1;
    }

    return ptm;
}

#ifdef USE_UTMPX
static const char *skip_prefix(const char *string, const char *prefix) {
    size_t length = strlen(prefix);

    if (!strncmp(string, prefix, length)) {
        return string + length;
    } else {
        return string;
    }
}
static int set_utmpx(short type, int ptm) {
    struct utmpx entry;

    const char *tty, *user, *id, *line;
    struct timeval now;

    user = getpwuid(getuid())->pw_name;
    gettimeofday(&now, NULL);

    tty = ptsname(ptm);
    if (!tty) {
        return 0;
    }

    id   = skip_prefix(tty, "/dev/pts/");
    line = skip_prefix(tty, "/dev/");

    /* Create utmp entry for the given terminal. */
    memset(&entry, 0, sizeof(entry));

    snprintf(entry.ut_user, sizeof(entry.ut_user), "%s", user);
    snprintf(entry.ut_id,   sizeof(entry.ut_id),   "%s", id);
    snprintf(entry.ut_line, sizeof(entry.ut_line), "%s", line);

    entry.ut_pid  = getpid();
    entry.ut_type = type;
    entry.ut_tv.tv_sec  = now.tv_sec;
    entry.ut_tv.tv_usec = now.tv_usec;

    /* Write the entry to the utmp file. */
    setutxent();
    if (!pututxline(&entry)) {
        return 0;
    }
    endutxent();

    return 1;
}
#endif
static int login(int ptm) {
#if defined(USE_UTEMPTER)
    return utempter_add_record(ptm, NULL);
#elif defined(USE_UTMPX)
    return set_utmpx(USER_PROCESS, ptm);
#else
# error "neither USE_UTEMPTER nor USE_UTMPX defined"
#endif
}
static int logout(int ptm) {
#if defined(USE_UTEMPTER)
    return utempter_remove_record(ptm);
#elif defined(USE_UTMPX)
    return set_utmpx(DEAD_PROCESS, ptm);
#else
# error "neither USE_UTEMPTER nor USE_UTMPX defined"
#endif
}

static void pass_buffer_to_program(const char *buffer, size_t length, char **argv) {
    int fds[2];
    FILE *fh;

    pid_t pid;

    /* Skip argv[0]. */
    argv++;

    if (pipe(fds) != 0) {
        perror("pipe");
        return;
    }

    fh = fdopen(fds[1] /* write side */, "w");
    if (!fh) {
        perror("fdopen");
        close(fds[0]);
        close(fds[1]);
        return;
    }

    pid = fork();
    if (pid < 0) {
        perror("fork");
        goto out;
    } else if (pid == 0) {
        /* child */

        close(fds[1]); /* write side */

        /* Pass read side as stdin to the program. */
        if (dup2(fds[0], STDIN_FILENO) < 0) {
            perror("dup2");
            exit(EXIT_FAILURE);
        }
        close(fds[0]);

        execvp(argv[0], argv);
        perror("execvp");
        exit(EXIT_FAILURE);
    }
    /* father */

    if (fwrite(buffer, 1, length, fh) != length) {
        perror("fwrite");
        /* continue to perform cleanup */
    }

out:
    close(fds[0]); /* read side */
    fclose(fh);
}
static void handle_wall(int fd, char **argv) {
    char buffer[4096];
    ssize_t r;

    while ((r = read(fd, buffer, sizeof(buffer))) > 0) {
        assert(SSIZE_MAX <= SIZE_MAX);
        pass_buffer_to_program(buffer, (size_t)r, argv);
    }
}


int main(int argc, char **argv) {
    int ptm, pts;
    char *name;

    if (argc < 2) {
        fprintf(stderr, "usage: %s <cmd args..>\n", argv[0]);
        exit(EXIT_FAILURE);
    }

    ptm = open_tty();
    if (ptm < 0) {
        perror("open_tty");
        exit(EXIT_FAILURE);
    }
    name = ptsname(ptm);
    if (!name) {
        perror("ptsname");
        exit(EXIT_FAILURE);
    }

#ifdef DEBUG
    printf("%s\n", name);
#endif

    /* We need to open the slave or reading from the master yields EOF after
     * the first wall write to it. */
    pts = open(name, O_RDWR);
    if (pts < 0) {
        perror(name);
        exit(EXIT_FAILURE);
    }

    /* Cleanup on signals. Necessary before login(). */
    setup_signals();

    if (!login(ptm)) {
        perror("login");
        exit(EXIT_FAILURE);
    }

    /* Main loop. Handle all wall messages sent to our PTY. */
    handle_wall(ptm, argv);

    if (!logout(ptm)) {
        perror("logout");
        exit(EXIT_FAILURE);
    }

    close(ptm);
    close(pts);

    return EXIT_SUCCESS;
}
